/*
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "hessian.h"
#include "i_hessian.h"
#include "linkedlist.h"
#include "log.h"


/**
 * Method prototypes
 */
static OBJECT_CTOR(hessian_list);
static OBJECT_DTOR(hessian_list);
static OBJECT_SERIALIZE(hessian_list);
static OBJECT_DESERIALIZE(hessian_list);

/**
 * Initializes and registers the Hessian list class.
 */
static const hessian_class_t _hessian_list_descr = {
    HESSIAN_LIST,
    "hessian.List",
    sizeof(hessian_list_t),
    'V', 'z',
    hessian_list_ctor,
    hessian_list_dtor,
    hessian_list_serialize,
    hessian_list_deserialize
};
const void * hessian_list_class = &_hessian_list_descr;


/**
 * Hessian list constructor. Creates an empty Hessian untyped list.
 *
 * hessian_object_t * list= hessian_create(HESSIAN_LIST);
 */
static hessian_object_t * hessian_list_ctor (hessian_object_t * list, va_list * ap) {
    hessian_list_t * self= list;
    if (self == NULL) {
        pep_log_error("hessian_list_ctor: NULL object pointer.");
        return NULL;
    }
    self->type= NULL;
    self->list= pep_llist_create();
    if (self->list == NULL) {
        pep_log_error("hessian_list_ctor: can't create list.");
        return NULL;
    }
    return self;
}


/**
 * Hessian list destructor. Recursively delete the contained objects.
 */
static int hessian_list_dtor (hessian_object_t * list) {
    hessian_list_t * self= list;
    if (self == NULL) {
        pep_log_error("hessian_list_dtor: NULL object pointer.");
        return HESSIAN_ERROR;
    }
    if (self->type != NULL) free(self->type);
    pep_llist_delete_elements(self->list,(pep_llist_delete_elt_f)hessian_delete);
    pep_llist_delete(self->list);
    return HESSIAN_OK;
}


/**
 * Adds an hessian_object_t * at the end of the list
 */
int hessian_list_add(hessian_object_t * list, hessian_object_t * object) {
    hessian_list_t * self= list;
    const hessian_class_t * class;
    if (self == NULL) {
        pep_log_error("hessian_list_add: NULL object pointer.");
        return HESSIAN_ERROR;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_add: NULL class descriptor.");
        return HESSIAN_ERROR;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_add: wrong class type: %d.",class->type);
        return HESSIAN_ERROR;
    }
    if (pep_llist_add(self->list, object) != LLIST_OK) {
        pep_log_error("hessian_list_add: can't add object to list.");
        return HESSIAN_ERROR;
    }
    return HESSIAN_OK;
}

/**
 * Hessian list serialize method.
 */
static int hessian_list_serialize (const hessian_object_t * list, pep_buffer_t * output) {
    const hessian_list_t * self= list;
    const hessian_class_t * class;
    size_t str_l, utf8_l, list_l;
    int b32, b24, b16, b8, i;
    if (self == NULL) {
        pep_log_error("hessian_list_serialize: NULL object pointer.");
        return HESSIAN_ERROR;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_serialize: NULL class descriptor.");
        return HESSIAN_ERROR;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_serialize: wrong class type: %d.",class->type);
        return HESSIAN_ERROR;
    }

    pep_buffer_putc(class->tag,output);
    /* write type if any */
    if (self->type != NULL) {
        str_l= strlen(self->type);
        utf8_l= hessian_utf8_strlen(self->type);
        b16= utf8_l >> 8;
        b8= utf8_l & 0x00FF;
        pep_buffer_putc('t',output);
        pep_buffer_putc(b16,output);
        pep_buffer_putc(b8,output);
        pep_buffer_write(self->type,1,str_l,output);
    }
    /* write length if any */
    list_l= pep_llist_length(self->list);
    if (list_l > 0) {
        int32_t value= (int32_t)list_l;
        b32 = (value >> 24) & 0x000000FF;
        b24 = (value >> 16) & 0x000000FF;
        b16 = (value >> 8) & 0x000000FF;
        b8 = value & 0x000000FF;
        pep_buffer_putc('l',output);
        pep_buffer_putc(b32,output);
        pep_buffer_putc(b24,output);
        pep_buffer_putc(b16,output);
        pep_buffer_putc(b8,output);
    }
    /* write all objects */
    i= 0;
    for( i= 0; i < list_l; i++ ) {
        hessian_object_t * object= pep_llist_get(self->list,i);
        if (object == NULL) {
            pep_log_error("hessian_list_add: NULL object pointer at: %d.",i);
            return HESSIAN_ERROR;
        }
        if (hessian_serialize(object, output) != HESSIAN_OK) {
            pep_log_error("hessian_list_add: can't serialize object at: %d.",i);
            return HESSIAN_ERROR;
        }
    }

    pep_buffer_putc(class->chunk_tag,output);
    return HESSIAN_OK;
}

/**
 * Hessian list deserialize method.
 */
static int hessian_list_deserialize (hessian_object_t * list, int tag, pep_buffer_t * input) {
    hessian_list_t * self= list;
    const hessian_class_t * class;
    pep_linkedlist_t * refs;
    size_t refs_l;
    int32_t length;
    int next_tag, i;
    if (self == NULL) {
        pep_log_error("hessian_list_deserialize: NULL object pointer.");
        return HESSIAN_ERROR;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_deserialize: NULL class descriptor.");
        return HESSIAN_ERROR;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_deserialize: wrong class type: %d.",class->type);
        return HESSIAN_ERROR;
    }
    /* tag is 'V' */
    if (tag != class->tag) {
        pep_log_error("hessian_list_deserialize: invalid tag: %c (%d).",(char)tag,tag);
        return HESSIAN_ERROR;
    }
    length= -1;
    /* alloc the refs list for Hessian ref handling */
    refs= pep_llist_create();
    if (refs == NULL) {
        pep_log_error("hessian_list_deserialize: can't create temp references list.");
        return HESSIAN_ERROR;
    }
    /* begin parsing */
    next_tag= pep_buffer_getc(input);
    /* optional type */
    if (next_tag == 't') {
        /* read the utf8 type length */
        int b16= pep_buffer_getc(input);
        int b8= pep_buffer_getc(input);
        size_t utf8_l= (b16 << 8) + b8;
        char * type= hessian_utf8_bgets(utf8_l,input);
        if (type == NULL) {
            pep_log_error("hessian_list_deserialize: can't read list type: %d chars.", (int)utf8_l);
            pep_llist_delete(refs);
            return HESSIAN_ERROR;
        }
        self->type= type;
        next_tag= pep_buffer_getc(input);
    }
    /* optional length, unused. */
    if (next_tag == 'l') {
        /* read int32, don't do anything with it... */
        int32_t b32 = pep_buffer_getc(input);
        int32_t b24 = pep_buffer_getc(input);
        int32_t b16 = pep_buffer_getc(input);
        int32_t b8 = pep_buffer_getc(input);
        length= (b32 << 24) + (b24 << 16) + (b16 << 8) + b8;
        next_tag= pep_buffer_getc(input);
    }
    /* do until tag != 'z' */
    while( next_tag != class->chunk_tag && next_tag != BUFFER_EOF) {
        /* standard Hessian object, add to the refs list. */
        hessian_object_t * o= hessian_deserialize_tag(next_tag,input);
        if (o == NULL) {
            pep_log_error("hessian_list_deserialize: can't deserialize object with tag: %c.", next_tag);
            pep_llist_delete_elements(refs,(pep_llist_delete_elt_f)hessian_delete);
            pep_llist_delete(refs);
            return HESSIAN_ERROR;
        }
        if (pep_llist_add(refs,o) != LLIST_OK) {
            pep_log_error("hessian_list_deserialize: can't add object to temp references list.");
            hessian_delete(o);
            pep_llist_delete_elements(refs,(pep_llist_delete_elt_f)hessian_delete);
            pep_llist_delete(refs);
            return HESSIAN_ERROR;
        }
        next_tag= pep_buffer_getc(input);
    }

    /* alloc the objects list and fill with element from the refs lists. */
    self->list= pep_llist_create();
    if (self->list == NULL) {
        pep_log_error("hessian_list_deserialize: can't create list.");
        pep_llist_delete_elements(refs,(pep_llist_delete_elt_f)hessian_delete);
        pep_llist_delete(refs);
        return HESSIAN_ERROR;
    }
    refs_l= pep_llist_length(refs);
    for (i= 0; i < refs_l; i++) {
        hessian_object_t * o= (hessian_object_t *)pep_llist_get(refs,i);
        if (o == NULL) {
            pep_log_error("hessian_list_deserialize: NULL object in refs at: %d.",i);
            pep_llist_delete_elements(refs,(pep_llist_delete_elt_f)hessian_delete);
            pep_llist_delete(refs);
            return HESSIAN_ERROR;
        }
        /* handle the ref object */
        if (hessian_gettype(o) == HESSIAN_REF) {
            /* get the ref index */
            int ref_index= hessian_ref_getvalue(o);
            hessian_delete(o); /* not needed anymore */
            /* get the real object */
            o= (hessian_object_t *)pep_llist_get(refs,ref_index);
            if (o == NULL) {
                pep_log_error("hessian_list_deserialize: NULL referenced object in refs at: %d.",ref_index);
                pep_llist_delete_elements(refs,(pep_llist_delete_elt_f)hessian_delete);
                pep_llist_delete(refs);
                return HESSIAN_ERROR;
            }
        }
        if (pep_llist_add(self->list,o) != LLIST_OK) {
            pep_log_error("hessian_list_deserialize: can't add object to list.");
            hessian_delete(o);
            pep_llist_delete_elements(refs,(pep_llist_delete_elt_f)hessian_delete);
            pep_llist_delete(refs);
            return HESSIAN_ERROR;
        }
    }
    pep_llist_delete(refs);
    return HESSIAN_OK;
}

/**
 * Sets the optional Hessian list type.
 */
int hessian_list_settype(hessian_object_t * list, const char * type) {
    hessian_list_t * self= list;
    const hessian_class_t * class;
    if (self == NULL) {
        pep_log_error("hessian_list_settype: NULL object pointer.");
        return HESSIAN_ERROR;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_settype: NULL class descriptor.");
        return HESSIAN_ERROR;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_settype: wrong class type: %d.",class->type);
        return HESSIAN_ERROR;
    }
    /* free if already set */
    if (self->type != NULL) {
        free(self->type);
        self->type= NULL;
    }
    if (type != NULL) {
        size_t type_l= strlen(type);
        self->type= calloc(type_l + 1, sizeof(char));
        if (self->type == NULL) {
            pep_log_error("hessian_list_settype: can't allocate type (%d chars).",(int)type_l);
            return HESSIAN_ERROR;
        }
        strncpy(self->type,type,type_l);
    }
    else {
        self->type= NULL;
    }
    return HESSIAN_OK;
}

/**
 * Returns the optional Hessian list type or NULL;
 */
const char * hessian_list_gettype(const hessian_object_t * list) {
    const hessian_list_t * self= list;
    const hessian_class_t * class;
    if (self == NULL) {
        pep_log_error("hessian_list_gettype: NULL object pointer.");
        return NULL;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_gettype: NULL class descriptor.");
        return NULL;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_gettype: wrong class type: %d.",class->type);
        return NULL;
    }
    return self->type;
}

/**
 * Returns the length of the Hessian list. returns 0 if empty or on error.
 */
size_t hessian_list_length(const hessian_object_t * list) {
    const hessian_list_t * self= list;
    const hessian_class_t * class;
    if (self == NULL) {
        pep_log_error("hessian_list_length: NULL object pointer.");
        return 0;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_length: NULL class descriptor.");
        return 0;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_length: wrong class type: %d.",class->type);
        return 0;
    }
    return pep_llist_length(self->list);
}

/**
 * Returns the Hessian object at index from the Hessian list. NULL if index > length.
 */
hessian_object_t * hessian_list_get(const hessian_object_t * list, int index) {
    const hessian_list_t * self= list;
    const hessian_class_t * class;
    if (self == NULL) {
        pep_log_error("hessian_list_get: NULL object pointer.");
        return NULL;
    }
    class= hessian_getclass(list);
    if (class == NULL) {
        pep_log_error("hessian_list_get: NULL class descriptor.");
        return NULL;
    }
    if (class->type != HESSIAN_LIST) {
        pep_log_error("hessian_list_get: wrong class type: %d.",class->type);
        return NULL;
    }
    return (hessian_object_t *) pep_llist_get(self->list,index);
}

